//
// File:        ArrayModule.java
// Package:     gov.llnl.babel.backend.fortran
// Copyright:   (c) 2003 The Regents of the University of California
// Release:     $Name: release-0-8-6 $
// Revision:    @(#) $Revision: 1.4 $
// Date:        $Date: 2003/03/19 19:00:02 $
// Description: Create a F90 module to provide stubs for arrays
// 
// Copyright (c) 2003, The Regents of the University of Calfornia.
// Produced at the Lawrence Livermore National Laboratory.
// Written by the Components Team <components@llnl.gov>
// UCRL-CODE-2002-054
// All rights reserved.
// 
// This file is part of Babel. For more information, see
// http://www.llnl.gov/CASC/components/. Please read the COPYRIGHT file
// for Our Notice and the LICENSE file for the GNU Lesser General Public
// License.
// 
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License (as published by
// the Free Software Foundation) version 2.1 dated February 1999.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
// conditions of the GNU Lesser General Public License for more details.
// 
// You should have recieved a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

package gov.llnl.babel.backend.fortran;
import gov.llnl.babel.BabelConfiguration;
import gov.llnl.babel.BabelConfiguration;
import gov.llnl.babel.backend.CodeGenerationException;
import gov.llnl.babel.backend.IOR;
import gov.llnl.babel.backend.mangler.NameMangler;
import gov.llnl.babel.backend.mangler.FortranMangler;
import gov.llnl.babel.backend.writers.LanguageWriter;
import gov.llnl.babel.symbols.SymbolID;
import gov.llnl.babel.symbols.Type;
import gov.llnl.babel.symbols.Version;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.List;
import java.security.NoSuchAlgorithmException;

/**
 * Create a FORTRAN 90 module to provide client-side bindings for arrays of 
 * classes and interfaces.
 */
public class ArrayModule {
  static private final int s_maxArray = BabelConfiguration.getMaximumArray();

  private NameMangler d_mang;

  private LanguageWriter d_lw;
  
  private SymbolID d_id;

  private String d_implArrayType;

  private String d_privateSuffix = "_p";

  private String d_intType;
  
  private boolean d_access;

  private boolean d_borrow;

  private boolean d_is_base_type;

  private static final String[] s_constant_methods = {
    "addRef",
    "copy",
    "create1d",
    "create2dCol",
    "create2dRow",
    "createCol",
    "createRow",
    "deleteRef",
    "dimen",
    "ensure",
    "isColumnOrder",
    "isRowOrder",
    "is_null",
    "lower",
    "not_null",
    "set_null",
    "slice",
    "smartCopy",
    "stride",
    "upper"
  };

  /**
   * The name of the type use to hold values going in/out.
   */
  private String d_implDataType;

  private SymbolID d_implID;

  private void initImplID(int type)
    throws CodeGenerationException
  {
    d_access = true;
    d_borrow = true;
    switch (type) {
    case Type.VOID:
      throw new CodeGenerationException("Cannot create an array for a void");
    case Type.CLASS:
    case Type.SYMBOL:
    case Type.INTERFACE:
      d_access = false;
      d_borrow = false;
      d_is_base_type = false;
      d_implID = d_id;
      break;
    case Type.ENUM:
      d_is_base_type = false;
      d_access = false;
      d_borrow = false;
      d_implID = new SymbolID("SIDL.int", new Version());
      break;
    case Type.CHAR:
    case Type.OPAQUE:
    case Type.STRING:
    case Type.BOOLEAN:
      d_access = false;
      d_borrow = false;
      // fall through intended!
    default:
      d_is_base_type = true;
      d_implID = new SymbolID("SIDL." + (new Type(type)).getTypeString(),
                              new Version());
      break;
    }
  }
    
  /**
   * Create an object to create a FORTRAN 90 module to provide
   * client-side bindings for arrays of classes, interfaces, and
   * enumerated types.
   *
   * @param id    the array module should be created for this
   *              id. This can be a real id or a fake id (for
   *              SIDL.double and the like).
   * @param type  this should be a constant from 
   *              {@link gov.llnl.babel.symbols.Type}. It indicates
   *              the type of id.
   * @param lw    here is where the file is created.
   * @exception gov.llnl.babel.backend.CodeGenerationException
   *  this means the arguments were incorrect
   */
  public ArrayModule(SymbolID id,
                     int      type,
                     LanguageWriter lw)
    throws CodeGenerationException
  {
    try {
      boolean isEnum = (Type.ENUM == type);
      d_id = id;
      d_lw = lw;
      d_mang = new FortranMangler(AbbrevHeader.MAXNAME,
                                  AbbrevHeader.MAXUNMANGLED);
      d_implArrayType = "type(" + Fortran.getArrayName(id) + ")";
      if (type <= Type.STRING) {
        d_implDataType = Fortran.getReturnString(new Type(type));
      }
      else {
        d_implDataType = Fortran.getReturnString(new Type(id));
      }
      d_intType = Fortran.getReturnString(new Type(Type.INT));
      initImplID(type);
    }
    catch (NoSuchAlgorithmException nsae) {
      throw new CodeGenerationException(nsae.getMessage());
    }
  }

  private void generateIsNull()
  {
    d_lw.println("logical function is_null" + d_privateSuffix +
                 "(array)");
    d_lw.increaseTabLevel();
    d_lw.println(d_implArrayType + ", intent(in) :: array");
    d_lw.println("is_null" + d_privateSuffix + " = (array%d_array .eq. 0)");
    d_lw.decreaseTabLevel();
    d_lw.println("end function is_null" + d_privateSuffix);
    d_lw.println();
  }

  private void generateNotNull()
  {
    d_lw.println("logical function not_null" + d_privateSuffix +
                 "(array)");
    d_lw.increaseTabLevel();
    d_lw.println(d_implArrayType + ", intent(in) :: array");
    d_lw.println("not_null" + d_privateSuffix + " = (array%d_array .ne. 0)");
    d_lw.decreaseTabLevel();
    d_lw.println("end function not_null" + d_privateSuffix);
    d_lw.println();
  }

  private void generateSetNull()
  {
    d_lw.println("subroutine set_null" + d_privateSuffix +
                 "(array)");
    d_lw.increaseTabLevel();
    d_lw.println(d_implArrayType + ", intent(out) :: array");
    d_lw.println("array%d_array = 0");
    d_lw.decreaseTabLevel();
    d_lw.println("end subroutine set_null" + d_privateSuffix);
    d_lw.println();
  }

  private void generateNullRelated()
  {
    generateIsNull();
    generateNotNull();
    generateSetNull();
  }

  private void generateCreate(String type)
    throws UnsupportedEncodingException
  {
    final String createFunc = "create" + type;
    final String function = d_mang.shortArrayName
      (d_implID.getFullName(), createFunc, "_m");
    d_lw.beginBlockComment(false);
    d_lw.println("The size of lower determines the dimension of the");
    d_lw.println("array.");
    d_lw.endBlockComment(false);
    d_lw.println("subroutine " + createFunc + d_privateSuffix +
                 "(lower, upper, array)");
    d_lw.increaseTabLevel();
    d_lw.println(d_intType + ", dimension(:), intent(in) :: lower");
    d_lw.println(d_intType +
                 ", dimension(:), intent(in) :: upper");
    d_lw.println(d_implArrayType + ", intent(out) :: array");
    d_lw.println(d_intType + " :: dim");
    d_lw.println("external " + function);
    d_lw.println("dim = size(lower)");
    d_lw.println("call " + function + "(dim, lower, upper, array)");
    d_lw.decreaseTabLevel();
    d_lw.println("end subroutine " + createFunc + d_privateSuffix);
    d_lw.println();
  }

  private void generateCreate1d()
    throws UnsupportedEncodingException
  {
    final String createFunc = "create1d";
    final String function = d_mang.shortArrayName
      (d_implID.getFullName(), createFunc, "_m");
    d_lw.println("subroutine " + createFunc + d_privateSuffix +
                 "(len, array)");
    d_lw.increaseTabLevel();
    d_lw.println(d_intType + ", intent(in) :: len");
    d_lw.println(d_implArrayType + ", intent(out) :: array");
    d_lw.println("external " + function);
    d_lw.println("call " + function + "(len, array)");
    d_lw.decreaseTabLevel();
    d_lw.println("end subroutine " + createFunc + d_privateSuffix);
    d_lw.println();
  }
  
  private void generateCreate2d(String type)
    throws UnsupportedEncodingException
  {
    final String createFunc = "create2d" + type;
    final String function = d_mang.shortArrayName
      (d_implID.getFullName(), createFunc, "_m");
    d_lw.println("subroutine " + createFunc + d_privateSuffix +
                 "(m, n, array)");
    d_lw.increaseTabLevel();
    d_lw.println(d_intType + ", intent(in) :: m, n");
    d_lw.println(d_implArrayType + ", intent(out) :: array");
    d_lw.println("external " + function);
    d_lw.println("call " + function + "(m, n, array)");
    d_lw.decreaseTabLevel();
    d_lw.println("end subroutine " + createFunc + d_privateSuffix);
    d_lw.println();
  }
  

  private void generateCreates()
    throws UnsupportedEncodingException
  {
    generateCreate("Col");
    generateCreate("Row");
    generateCreate1d();
    generateCreate2d("Col");
    generateCreate2d("Row");
  }

  private void generateCopy(String func)
    throws UnsupportedEncodingException
  {
    final String function = d_mang.shortArrayName
      (d_implID.getFullName(), func, "_m");
    d_lw.println("subroutine " + func + d_privateSuffix +
                 "(src, dest)");
    d_lw.increaseTabLevel();
    d_lw.println(d_implArrayType + ", intent(in) :: src");
    d_lw.println(d_implArrayType + ", intent(in) :: dest");
    d_lw.println("external " + function);
    d_lw.println("call " + function + "(src, dest)");
    d_lw.decreaseTabLevel();
    d_lw.println("end subroutine " + func + d_privateSuffix);
    d_lw.println();
  }

  private void generateCopies()
    throws UnsupportedEncodingException
  {
    generateCopy("copy");
    generateCopy("smartCopy");
  }

  private void generateEnsure()
    throws UnsupportedEncodingException
  {
    final String function = d_mang.shortArrayName
      (d_implID.getFullName(), "ensure", "_m");
    d_lw.println("subroutine ensure" + d_privateSuffix +
                 "(src, dim, ordering, result)");
    d_lw.increaseTabLevel();
    d_lw.println(d_implArrayType + ", intent(in)  :: src");
    d_lw.println(d_implArrayType + ", intent(out) :: result");
    d_lw.println(d_intType + ", intent(in) :: dim, ordering");
    d_lw.println("external " + function);
    d_lw.println("call " + function + 
                 "(src, dim, ordering, result)");
    d_lw.decreaseTabLevel();
    d_lw.println("end subroutine ensure" + d_privateSuffix);
    d_lw.println();
  }

  private void generateSlice()
    throws UnsupportedEncodingException
  {
    final String function = d_mang.shortArrayName
      (d_implID.getFullName(), "slice", "_m");
    d_lw.println("subroutine slice" + d_privateSuffix +
                 "(src, dim, numElem, srcStart, srcStride, newLower, result)");
    d_lw.increaseTabLevel();
    d_lw.println(d_implArrayType + ", intent(in)  :: src");
    d_lw.println(d_intType + ", intent(in) :: dim");
    d_lw.println(d_intType + ", dimension(:), intent(in) :: numElem");
    d_lw.println
      (d_intType + 
       ", dimension(:), intent(in) :: srcStart, srcStride");
    d_lw.println(d_implArrayType + ", intent(out) :: result");
    d_lw.println(d_intType + ", dimension(:), intent(in) :: newLower");
    d_lw.println("external " + function);
    d_lw.println("call " + function + 
                 "(src, dim, numElem, srcStart, srcStride, newLower, result)");
    d_lw.decreaseTabLevel();
    d_lw.println("end subroutine slice" + d_privateSuffix);
    d_lw.println();
  }

  private void generateQuery(String func, String returnType)
    throws UnsupportedEncodingException
  {
    final String function = d_mang.shortArrayName
      (d_implID.getFullName(), func, "_m");
    d_lw.println(returnType + " function  " + func + d_privateSuffix +
                 "(array)");
    d_lw.increaseTabLevel();
    d_lw.println(d_implArrayType + ", intent(in)  :: array");
    d_lw.println("external " + function);
    d_lw.println("call " + function + 
                 "(array, " + func + d_privateSuffix + ")");
    d_lw.decreaseTabLevel();
    d_lw.println("end function " + func + d_privateSuffix);
    d_lw.println();
  }

  private void generateOrderingTests()
    throws UnsupportedEncodingException
  {
    generateQuery("isColumnOrder", "logical");
    generateQuery("isRowOrder", "logical");
  }

  private void generateIndexedQuery(String func)
    throws UnsupportedEncodingException
  {
    final String function = d_mang.shortArrayName
      (d_implID.getFullName(), func, "_m");
    d_lw.println(d_intType + " function  " + func + d_privateSuffix +
                 "(array, index)");
    d_lw.increaseTabLevel();
    d_lw.println(d_implArrayType + ", intent(in)  :: array");
    d_lw.println(d_intType + ", intent(in) :: index");
    d_lw.println("external " + function);
    d_lw.println("call " + function + 
                 "(array, index, " + func + d_privateSuffix + ")");
    d_lw.decreaseTabLevel();
    d_lw.println("end function " + func + d_privateSuffix);
    d_lw.println();
  }

  private void generateSizeQueries()
    throws UnsupportedEncodingException
  {
    generateQuery("dimen", d_intType);
    generateIndexedQuery("stride");
    generateIndexedQuery("lower");
    generateIndexedQuery("upper");
  }

  private void generateGenericGet()
    throws UnsupportedEncodingException
  {
    final String function = d_mang.shortArrayName
      (d_implID.getFullName(), "get", "_m");
    //    d_lw.println(d_implDataType + " function get" + d_privateSuffix +
    //                 "(array, index)");
    d_lw.println("subroutine get" + d_privateSuffix +
                 "(array, index, value)");
    d_lw.increaseTabLevel();
    d_lw.println(d_implArrayType + ", intent(in)  :: array");
    d_lw.println(d_intType + ", intent(in), dimension(:) :: index");
    d_lw.println(d_implDataType + ", intent(out) :: value");
    d_lw.println("external " + function);
    d_lw.println("call " + function + 
                 "(array, index, value)");
    d_lw.decreaseTabLevel();
    d_lw.println("end subroutine get" + d_privateSuffix);
    d_lw.println();
  }

  private void generateGet(int dim)
    throws UnsupportedEncodingException
  {
    final String func = "get" + Integer.toString(dim);
    int i;
    final String function = d_mang.shortArrayName
      (d_implID.getFullName(), func, "_m");
    d_lw.println("subroutine " + func + d_privateSuffix +
                 "(array, &");
    d_lw.increaseTabLevel();
    d_lw.increaseTabLevel();
    for(i = 1; i <= dim; ++i){
      d_lw.println("i" + i + ", &");
    }
    d_lw.println("value)");
    d_lw.decreaseTabLevel();
    d_lw.println(d_implArrayType + ", intent(in)  :: array");
    for(i = 1; i <= dim; ++i) {
      d_lw.println(d_intType + ", intent(in) :: i" + i);
    }
    d_lw.println(d_implDataType + ", intent(out) :: value");
    d_lw.println("external " + function);
    d_lw.println("call " + function + "(array, &");
    d_lw.increaseTabLevel();
    for(i = 1; i <= dim; ++i) {
      d_lw.println("i" + i + ", &");
    }
    d_lw.println("value)");
    d_lw.decreaseTabLevel();
    d_lw.decreaseTabLevel();
    d_lw.println("end subroutine " + func + d_privateSuffix);
    d_lw.println();
  }

  private void generateGenericSet()
    throws UnsupportedEncodingException
  {
    final String function = d_mang.shortArrayName
      (d_implID.getFullName(), "set", "_m");
    d_lw.println("subroutine set" + d_privateSuffix +
                 "(array, index, value)");
    d_lw.increaseTabLevel();
    d_lw.println(d_implArrayType + ", intent(in)  :: array");
    d_lw.println(d_intType + ", intent(in), dimension(:) :: index");
    d_lw.println(d_implDataType + ", intent(in) :: value");
    d_lw.println("external " + function);
    d_lw.println("call " + function + 
                 "(array, index, value)");
    d_lw.decreaseTabLevel();
    d_lw.println("end subroutine set" + d_privateSuffix);
    d_lw.println();
  }

  private void generateSet(int dim)
    throws UnsupportedEncodingException
  {
    final String func = "set" + Integer.toString(dim);
    int i;
    final String function = d_mang.shortArrayName
      (d_implID.getFullName(), func, "_m");
    d_lw.println("subroutine " + func + d_privateSuffix +
                 "(array, &");
    d_lw.increaseTabLevel();
    d_lw.increaseTabLevel();
    for(i = 1; i <= dim; ++i){
      d_lw.println("i" + i + ", &");
    }
    d_lw.println("value)");
    d_lw.decreaseTabLevel();
    d_lw.println(d_implArrayType + ", intent(in)  :: array");
    for(i = 1; i <= dim; ++i) {
      d_lw.println(d_intType + ", intent(in) :: i" + i);
    }
    d_lw.println(d_implDataType + ", intent(in) :: value");
    d_lw.println("external " + function);
    d_lw.println("call " + function + 
                 "(array, &");
    d_lw.increaseTabLevel();
    for(i = 1; i <= dim; ++i) {
      d_lw.println("i" + i + ", &");
    }
    d_lw.println("value)");
    d_lw.decreaseTabLevel();
    d_lw.decreaseTabLevel();
    d_lw.println("end subroutine " + func + d_privateSuffix);
    d_lw.println();
  }

  private void generateGetSet()
    throws UnsupportedEncodingException
  {
    int i;
    final int maxarray = BabelConfiguration.getMaximumArray();
    generateGenericGet();
    generateGenericSet();
    for(i = 1; i <= maxarray; ++i){
      generateGet(i);
      generateSet(i);
    }
  }

  private void writeInterface(String method)
  {
    int i;
    final int maxarray = BabelConfiguration.getMaximumArray();
    d_lw.print("private :: " + method + d_privateSuffix);
    for(i = 1; i <= maxarray; ++i) {
      d_lw.print(", " + method + Integer.toString(i) + d_privateSuffix);
    }
    d_lw.println();
    d_lw.println("interface " + method);
    d_lw.increaseTabLevel();
    d_lw.print("module procedure " + method + d_privateSuffix);
    for(i = 1; i <= maxarray; ++i) {
      d_lw.print(", " + method + Integer.toString(i) + d_privateSuffix);
    }
    d_lw.println();
    d_lw.decreaseTabLevel();
    d_lw.println("end interface");
    d_lw.println();
  }

  private void writeAccessInterface()
  {
    d_lw.println("private :: access" + d_privateSuffix);
    d_lw.println();
    d_lw.println("interface access");
    d_lw.increaseTabLevel();
    d_lw.println("module procedure access" + d_privateSuffix);
    d_lw.decreaseTabLevel();
    d_lw.println("end interface");
    d_lw.println();
  }

  private void generateAccess()
    throws UnsupportedEncodingException
  {
    final String function = d_mang.shortArrayName
      (d_implID.getFullName(), "access", "_m");
    d_lw.println("subroutine access" + d_privateSuffix +
                 "(array, ref, low, up, str, index)");
    d_lw.increaseTabLevel();
    d_lw.println("implicit none");
    d_lw.println(d_implArrayType + ", intent(in) :: array");
    d_lw.println(d_implDataType + ", dimension(1), intent(in) :: ref");
    d_lw.println(d_intType + ", intent(out), dimension(:) :: low, up, str");
    d_lw.println(d_intType + ", intent(out) :: index");
    d_lw.println("external " + function);
    d_lw.println("call " + function +
                 "(array, ref, low, up, str, index)");
    d_lw.decreaseTabLevel();
    d_lw.println("end subroutine access" + d_privateSuffix);
  }

  private void writeBorrowInterface()
  {
    d_lw.println("private :: borrow" + d_privateSuffix);
    d_lw.println();
    d_lw.println("interface borrow");
    d_lw.increaseTabLevel();
    d_lw.println("module procedure borrow" + d_privateSuffix);
    d_lw.decreaseTabLevel();
    d_lw.println("end interface");
    d_lw.println();
  }

  private void generateBorrow()
    throws UnsupportedEncodingException
  {
    final String func = "borrow";
    final String function = d_mang.shortArrayName
      (d_implID.getFullName(), func, "_m");
    d_lw.println("subroutine " + func + d_privateSuffix +
                 "(first, dim, low, up, astr, result)");
    d_lw.increaseTabLevel();
    d_lw.println(d_implDataType + ", intent(in) :: first");
    d_lw.println(d_intType + ", intent(in) :: dim");
    d_lw.println(d_intType + ", dimension(:), intent(in) :: low, up, astr");
    d_lw.println(d_implArrayType + ", intent(out) :: result");
    d_lw.println("external " + function);
    d_lw.println("call " + function + "(first, dim, low, up, astr, result)");
    d_lw.decreaseTabLevel();
    d_lw.println("end subroutine " + func + d_privateSuffix);
    d_lw.println();
  }

  private void generateDeclarations()
    throws UnsupportedEncodingException
  {
    int i;
    if (d_is_base_type) {
      TypeModule.writeArrayType(d_lw, d_id);
    }
    for(i = 0; i < s_constant_methods.length ; ++i){
      d_lw.println("private :: " + s_constant_methods[i] + 
                   d_privateSuffix);
      d_lw.println("interface " + s_constant_methods[i]);
      d_lw.increaseTabLevel();
      d_lw.println("module procedure " + s_constant_methods[i] + 
                   d_privateSuffix);
      d_lw.decreaseTabLevel();
      d_lw.println("end interface");
      d_lw.println();
    }
    writeInterface("get");
    writeInterface("set");
    if (d_access) {
      writeAccessInterface();
    }
    if (d_borrow) {
      writeBorrowInterface();
    }
  }

  private void generateRef(String func)
    throws UnsupportedEncodingException
  {
    final String function = d_mang.shortArrayName
      (d_implID.getFullName(), func, "_m");
    d_lw.println("subroutine  " + func + d_privateSuffix +
                 "(array)");
    d_lw.increaseTabLevel();
    d_lw.println(d_implArrayType + ", intent(in)  :: array");
    d_lw.println("external " + function);
    d_lw.println("call " + function + 
                 "(array)");
    d_lw.decreaseTabLevel();
    d_lw.println("end subroutine " + func + d_privateSuffix);
    d_lw.println();
  }

  private void generateRefCounting()
    throws UnsupportedEncodingException
  {
    generateRef("addRef");
    generateRef("deleteRef");
  }

  private void generateImplementations()
    throws UnsupportedEncodingException
  {
    generateCreates();
    generateCopies();
    generateEnsure();
    generateSlice();
    generateOrderingTests();
    generateSizeQueries();
    generateGetSet();
    generateRefCounting();
    generateNullRelated();
    if (d_access) {
      generateAccess();
    }
    if (d_borrow) {
      generateBorrow();
    }
  }

  public void generateStub()
    throws CodeGenerationException
  {
    try {
      if (!d_is_base_type) {
        d_lw.printlnUnformatted("#include \"" +
                                Fortran.getStubNameFile(d_id) +
                                "\"");
      }
      d_lw.println("module " + Fortran.getArrayModule(d_id));
      d_lw.increaseTabLevel();
      if (!d_is_base_type) {
        d_lw.println("use " + Fortran.getTypeModule(d_id));
      }
      d_lw.println();
      
      generateDeclarations();

      d_lw.println();
      d_lw.decreaseTabLevel();
      d_lw.println("contains");
      d_lw.increaseTabLevel();
      d_lw.println();
      d_lw.println();

      generateImplementations();
      d_lw.decreaseTabLevel();
      d_lw.println("end module " +  Fortran.getArrayModule(d_id));
    }
    catch (UnsupportedEncodingException uee) {
      throw new CodeGenerationException(uee.getMessage());
    }
  }
}
