//
// File:        DTDManager.java
// Package:     gov.llnl.babel.parsers.xml
// Release:     $Name: release-0-8-6 $
// Revision:    @(#) $Id: DTDManager.java,v 1.11 2003/06/03 21:02:12 epperly Exp $
// Description: singleton manager for SIDL symbol DTDs
//
// Copyright (c) 2000-2001, The Regents of the University of Calfornia.
// Produced at the Lawrence Livermore National Laboratory.
// Written by the Components Team <components@llnl.gov>
// UCRL-CODE-2002-054
// All rights reserved.
// 
// This file is part of Babel. For more information, see
// http://www.llnl.gov/CASC/components/. Please read the COPYRIGHT file
// for Our Notice and the LICENSE file for the GNU Lesser General Public
// License.
// 
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License (as published by
// the Free Software Foundation) version 2.1 dated February 1999.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
// conditions of the GNU Lesser General Public License for more details.
// 
// You should have recieved a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

package gov.llnl.babel.parsers.xml;

import gov.llnl.babel.Version;
import gov.llnl.babel.url.URLUtilities;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.StringTokenizer;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;

/**
 * <code>DTDManager</code> is a singleton class that manages the DTDs for
 * the SIDL symbol, comment, and HTML descriptions needed to parse SIDL
 * symbols in XML format.  If the application does not explicitly set the
 * path to the SIDL DTDs, the the DTD manager will query the system property
 * "gov.llnl.babel.SIDL.DTDPath".  This manager also defines the PUBLIC
 * identifiers for symbol and comment DTDs used to validate SIXL XML symbols.
 * The PUBLIC symbol identifier is "-//CCA/SIDL Symbol DTD v1.0//EN" and the
 * HTML lite comment identifier is "-//CCA/SIDL HTML DTD v1.0//EN".  The DTD
 * manager object can act as a symbol resolver for a SAX parser to interpret
 * these PUBLIC identifiers.
 */
public class DTDManager implements EntityResolver {
  public static final String SYMBOL_PUBLIC_ID =
    "-//CCA//SIDL Symbol DTD v1.1//EN";
  public static final String HTML_PUBLIC_ID =
    "-//CCA//SIDL HTML DTD v1.0//EN";
  public static final String COMMENT_PUBLIC_ID =
    "-//CCA//SIDL Comment DTD v1.0//EN";
  public static final String COMMENT_FILE = "comment.dtd";
  private static final String s_resource_id =
    "gov/llnl/babel/dtds/";

   private static DTDManager s_instance = null;

  JarFile d_jarFile = null;

   /**
    * Create a new instance of the DTD manager.  Although singleton classes
    * do not typically define a public constructor, this implementation does
    * so to support multiple DTD managers in the same application.  Most
    * implementations, however, will not directly create a DTD manager
    * through the constructor and will instead use the singleton functions
    * <code>getInstance</code> and <code>setInstance</code>.
    */
   public DTDManager() {
   }

   /**
    * Return the singleton instance of the DTD manager.  If the symbol
    * table instance has not yet been created, then it will be created
    * by this call.
    */
   public static DTDManager getInstance() {
      if (s_instance == null) {
         s_instance = new DTDManager();
      }
      return s_instance;
   }

   /**
    * Set the singleton instance of the DTD manager.  Use this function
    * only if you want to change the default DTD manager implementation
    * or use multiple DTD managers in the same application.  Otherwise,
    * <code>getInstance</code> will automatically create the DTD manager
    * for you.
    */
   public static void setInstance(DTDManager instance) {
      s_instance = instance;
   }

   /**
    * Provide the implementation of an <code>EntityResolver</code> to
    * convert the symbol PUBLIC identifier into a URL for the symbol DTD.
    */
  public InputSource resolveEntity(String public_id, String system_id) 
    throws java.io.IOException
  {
    /*
     * Check whether any of the public IDs match SIDL public IDs.
     */
    InputSource is = null;
    String jarEntry = null;
    String simpleName = null;
    if (SYMBOL_PUBLIC_ID.equals(public_id)) {
      simpleName = "SIDL.dtd";
      jarEntry = s_resource_id + simpleName;
    }
    else if (HTML_PUBLIC_ID.equals(public_id)) {
      simpleName = "html-lite.dtd";
      jarEntry = s_resource_id + simpleName;
    }
    else if (COMMENT_PUBLIC_ID.equals(public_id)) {
      simpleName = "comment.dtd";
      jarEntry = s_resource_id  + simpleName;
    }
    if (jarEntry != null) {
      ClassLoader loader = getClass().getClassLoader();
      if (loader == null) {
        return searchClassPath(jarEntry);
      }
      InputStream in = loader.getResourceAsStream(jarEntry);
      if (in != null) {
        is = new InputSource(in);
      }
      else {
        throw new IOException("Can't load " + simpleName + 
                              " from jar file");
      }
    } 
    return is;
  }

  private InputSource searchClassPath(String name)
    throws java.io.IOException
  {
    if (d_jarFile == null) {
      final String babelJar = "babel-" + Version.VERSION + ".jar";
      String classpath = System.getProperty("java.class.path", "");
      StringTokenizer tok = new StringTokenizer(classpath, 
                                                File.pathSeparator);
      while (tok.hasMoreTokens()) {
        String path = tok.nextToken();
        if (path.endsWith(babelJar)) {
          File f = new File(path);
          if (f.isFile()) {
            try {
              JarFile jar = new JarFile(f);
              if (jar != null) {
                d_jarFile = jar;
                break;
              }
            }
            catch (java.io.IOException ioe) {
              // don't do anything skip to next jar file
            }
          }
        }
      }
    }
    if (d_jarFile != null) {
      JarEntry entry = d_jarFile.getJarEntry(name);
      InputStream is = d_jarFile.getInputStream(entry);
      if (is != null) {
        return new InputSource(is);
      }
    }
    throw new IOException("Cannot get a ClassLoader to load XML DTDs");
  }
}
